#!/usr/bin/env python3
"""Utilities to localize and analyze orbitals."""
import numpy as np
from scipy.linalg import eig, expm, norm
from scipy.stats import unitary_group

from .logger import log
from .utils import handle_spin_gracefully


def eval_psi(atoms, psi, r):
    """Evaluate orbitals at given coordinate points.

    Args:
        atoms: Atoms object.
        psi (ndarray): Set of orbitals in reciprocal space.
        r (ndarray): Real-space positions.

    Returns:
        ndarray: Values of psi at points r.
    """
    # Shift the evaluation point to (0,0,0) because we always have a lattice point here
    psi_T = atoms.T(psi, -r)
    psi_Trs = atoms.I(psi_T)
    # The zero entry is always the value at point (0,0,0)
    return psi_Trs[0]


def get_R(atoms, psi, fods):
    """Calculate transformation matrix to build Fermi orbitals.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        atoms: Atoms object.
        psi (ndarray): Set of orbitals in reciprocal space.
        fods (list): Fermi-orbital descriptors.

    Returns:
        ndarray: Transformation matrix R.
    """
    # We only calculate occupied orbitals
    R = np.empty((len(fods), len(fods)), dtype=complex)

    for i in range(len(fods)):
        # Get the value at one FOD position for all psi
        psi_fod = eval_psi(atoms, psi, fods[i])
        sum_psi_fod = np.sqrt(np.sum(psi_fod.conj() * psi_fod))
        for j in range(len(fods)):
            R[i, j] = psi_fod[j].conj() / sum_psi_fod
    return R


def get_FO(atoms, psi, fods):
    """Calculate Fermi orbitals from Kohn-Sham orbitals.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        atoms: Atoms object.
        psi (ndarray): Set of orbitals in reciprocal space.
        fods (list): Fermi-orbital descriptors.

    Returns:
        ndarray: Real-space Fermi orbitals.
    """
    fo = np.zeros((atoms.occ.Nspin, atoms.Ns, atoms.occ.Nstate), dtype=complex)

    # Transform psi to real-space
    psi_rs = atoms.I(psi)
    for spin in range(atoms.occ.Nspin):
        # Get the transformation matrix R
        R = get_R(atoms, psi[spin], fods[spin])
        for i in range(len(R)):
            for j in range(atoms.occ.Nstate):
                if atoms.occ.f[spin, j] > 0:
                    fo[spin, :, i] += R[i, j] * psi_rs[spin, :, j]
    return fo


@handle_spin_gracefully
def get_S(atoms, psirs):
    """Calculate overlap matrix between orbitals.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        ndarray: Overlap matrix S.
    """
    # Overlap elements: S_ij = \int psi_i^* psi_j dr
    S = np.empty((atoms.occ.Nstate, atoms.occ.Nstate), dtype=complex)

    for i in range(atoms.occ.Nstate):
        for j in range(atoms.occ.Nstate):
            S[i, j] = atoms.dV * np.sum(psirs[:, i].conj() * psirs[:, j])
    return S


def get_FLO(atoms, psi, fods):
    """Calculate Fermi-Loewdin orbitals by orthonormalizing Fermi orbitals.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        atoms: Atoms object.
        psi (ndarray): Set of orbitals in reciprocal space.
        fods (list): Fermi-orbital descriptors.

    Returns:
        ndarray: Real-space Fermi-Loewdin orbitals.
    """
    fo = get_FO(atoms, psi, fods)
    flo = np.empty((atoms.occ.Nspin, atoms.Ns, atoms.occ.Nstate), dtype=complex)

    for spin in range(atoms.occ.Nspin):
        # Calculate the overlap matrix for FOs
        S = get_S(atoms, fo[spin])
        # Calculate eigenvalues and eigenvectors
        Q, T = eig(S)
        # Loewdins symmetric orthonormalization method
        Q12 = np.diag(1 / np.sqrt(Q))
        flo[spin] = fo[spin] @ (T @ Q12 @ T.T)
    return flo


@handle_spin_gracefully
def wannier_cost(atoms, psirs):
    """Calculate the Wannier cost function, namely the orbital variance. Equivalent to Foster-Boys.

    This function does not account for periodicity, it may be a good idea to center the system.

    Reference: J. Chem. Phys. 137, 224114.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        ndarray: Variance per orbital.
    """
    # Variance = \int psi r^2 psi - (\int psi r psi)^2
    centers = wannier_center(atoms, psirs)
    moments = second_moment(atoms, psirs)
    costs = moments - norm(centers, axis=1)**2
    log.debug(f'Centers:\n{centers}\nMoments:\n{moments}')
    log.info(f'Costs:\n{costs}')
    return costs


@handle_spin_gracefully
def wannier_center(atoms, psirs):
    """Calculate Wannier centers, i.e., the expectation values of r.

    Reference: J. Chem. Phys. 137, 224114.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        ndarray: Wannier centers per orbital.
    """
    centers = np.empty((atoms.occ.Nstate, 3))
    for i in range(atoms.occ.Nstate):
        for dim in range(3):
            centers[i, dim] = atoms.dV * np.real(np.sum(psirs[:, i].conj() * atoms.r[:, dim] *
                                                 psirs[:, i], axis=0))
    return centers


@handle_spin_gracefully
def second_moment(atoms, psirs):
    """Calculate the second moments, i.e., the expectation values of r^2.

    Reference: J. Chem. Phys. 137, 224114.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        ndarray: Second moments per orbital.
    """
    r2 = norm(atoms.r, axis=1)**2

    moments = np.empty(atoms.occ.Nstate)
    for i in range(atoms.occ.Nstate):
        moments[i] = atoms.dV * np.real(np.sum(psirs[:, i].conj() * r2 * psirs[:, i], axis=0))
    return moments


@handle_spin_gracefully
def wannier_supercell_matrices(atoms, psirs):
    """Calculate matrices for the supercell Wannier localization.

    Reference: Phys. Rev. B 59, 9703.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        tuple[ndarray, ndarray, ndarray]: Matrices X, Y, and Z.
    """
    # Similar to the expectation value of r, but accounting for periodicity
    X = (psirs.conj().T * np.exp(-1j * 2 * np.pi * atoms.r[:, 0] / atoms.a[0])) @ psirs
    Y = (psirs.conj().T * np.exp(-1j * 2 * np.pi * atoms.r[:, 1] / atoms.a[1])) @ psirs
    Z = (psirs.conj().T * np.exp(-1j * 2 * np.pi * atoms.r[:, 2] / atoms.a[2])) @ psirs
    return X * atoms.dV, Y * atoms.dV, Z * atoms.dV


def wannier_supercell_cost(X, Y, Z):
    """Calculate the supercell Wannier cost.

    This is an equivalent criterion to the spread criterion, but not the same. This cost function
    will be maximized instead of the minimization of the spread.

    Reference: Phys. Rev. B 59, 9703.

    Args:
        X (ndarray): Calculation specific matrix.
        Y (ndarray): Calculation specific matrix.
        Z (ndarray): Calculation specific matrix.

    Returns:
        float: Supercell Wannier cost.
    """
    X2 = np.abs(np.diagonal(X))**2
    Y2 = np.abs(np.diagonal(Y))**2
    Z2 = np.abs(np.diagonal(Z))**2
    return np.sum(X2 + Y2 + Z2)


def wannier_supercell_grad(atoms, X, Y, Z):
    """Calculate the supercell Wannier gradient.

    Reference: Phys. Rev. B 59, 9703.

    Args:
        atoms: Atoms object.
        X (ndarray): Calculation specific matrix.
        Y (ndarray): Calculation specific matrix.
        Z (ndarray): Calculation specific matrix.

    Returns:
        ndarray: Supercell Wannier gradient.
    """
    x = np.zeros((atoms.occ.Nstate, atoms.occ.Nstate), dtype=complex)
    y = np.zeros((atoms.occ.Nstate, atoms.occ.Nstate), dtype=complex)
    z = np.zeros((atoms.occ.Nstate, atoms.occ.Nstate), dtype=complex)
    # Just the indexed gradient from the paper, without fancy optimization
    for n in range(atoms.occ.Nstate):
        for m in range(atoms.occ.Nstate):
            x[m, n] = X[n, m] * (X[n, n].conj() - X[m, m].conj()) \
                - X[m, n].conj() * (X[m, m] - X[n, n])
            y[m, n] = Y[n, m] * (Y[n, n].conj() - Y[m, m].conj()) \
                - Y[m, n].conj() * (Y[m, m] - Y[n, n])
            z[m, n] = Z[n, m] * (Z[n, n].conj() - Z[m, m].conj()) \
                - Z[m, n].conj() * (Z[m, m] - Z[n, n])
    return x + y + z


@handle_spin_gracefully
def get_wannier(atoms, psirs, Nit=10000, conv_tol=1e-7, mu=0.25, random_guess=False, seed=None):
    """Steepest descent supercell Wannier localization.

    This function is rather sensitive to the starting point, thus it is a good idea to start from
    already localized orbitals.

    This optimizes the given orbitals under unitary constraint matrices, see
    IEEE Trans. Signal Process. 56, 1134.

    Reference: Phys. Rev. B 59, 9703.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Keyword Args:
        Nit (int): Number of iterations.
        conv_tol (float): Convergence tolerance.
        mu (float): Step size.
        random_guess (bool): Weather to use a random unitary starting guess or the identity.
        seed (int | None): Seed to get a reproducible random guess.

    Returns:
        ndarray: Localized orbitals.
    """
    X, Y, Z = wannier_supercell_matrices(atoms, psirs)  # Calculate matrices only once
    # The initial unitary transformation is the identity or a random unitary matrix
    if random_guess and atoms.occ.Nstate > 1:
        U = unitary_group.rvs(atoms.occ.Nstate, random_state=seed)
    else:
        U = np.eye(atoms.occ.Nstate)
    costs = [0]  # Add a zero to the costs to allow the sign evaluation in the first iteration

    atoms.log.debug(f'{"Iteration":<11}{"Cost [a0^2]":<13}{"dCost [a0^2]":<13}')
    for i in range(Nit):
        sign = 1
        costs.append(wannier_supercell_cost(X, Y, Z))
        if abs(costs[-2] - costs[-1]) < conv_tol:
            atoms.log.info(f'Wannier localizer converged after {i} iterations.')
            break
        # If the cost function gets smaller, change the direction
        if costs[-2] - costs[-1] < 0:
            sign = -1

        # Calculate unitary transformation
        dOmega = wannier_supercell_grad(atoms, X, Y, Z)
        A = sign * mu * dOmega
        # dOmega is anti-hermitian, therefore calculate -A instead of A.conj().T
        # expm(A) will be unitary
        expA_pos, expA_neg = expm(A), expm(-A)
        # Update total rotation
        U = U @ expA_pos
        # Update matrices
        X = expA_neg @ X @ expA_pos
        Y = expA_neg @ Y @ expA_pos
        Z = expA_neg @ Z @ expA_pos

        atoms.log.debug(f'{i:>8}   {costs[-1]:<+13,.6f}{costs[-2] - costs[-1]:<+13,.4e}')

    if len(costs) > 1 and abs(costs[-2] - costs[-1]) > conv_tol:
        atoms.log.warning('Wannier localizer not converged!')
    # Return the localized orbitals by rotating them
    return psirs @ U



#!/usr/bin/env python3
"""PDB file handling."""
import numpy as np

from ..logger import log
from ..units import bohr2ang


def write_pdb(obj, filename, fods=None, elec_symbols=('X', 'He'), trajectory=False):
    """Generate PDB files from atoms objects.

    See :func:`~eminus.io.pdb.create_pdb_str` for more information about the PDB file format.

    Args:
        obj: Atoms or SCF object.
        filename (str): PDB output file path/name.

    Keyword Args:
        fods (list): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.
        trajectory (bool): Allow appending to a file to create trajectories.

    Returns:
        None.
    """
    atoms = obj._atoms

    if not filename.endswith('.pdb'):
        filename += '.pdb'

    if 'He' in atoms.atom and atoms.unrestricted:
        log.warning('You need to modify "elec_symbols" to write helium with FODs in the spin-'
                    'polarized case.')

    atom = atoms.atom
    X = atoms.X
    if fods is not None:
        if len(fods[0]) != 0:
            atom = atom + [elec_symbols[0]] * len(fods[0])
            X = np.vstack((X, fods[0]))
        if len(fods) > 1 and len(fods[1]) != 0:
            atom = atom + [elec_symbols[1]] * len(fods[1])
            X = np.vstack((X, fods[1]))

    # Append to a file when using the trajectory keyword
    if trajectory:
        mode = 'a'
    else:
        mode = 'w'

    with open(filename, mode) as fp:
        fp.write(create_pdb_str(atom, X, a=atoms.a))


def create_pdb_str(atom, X, a=None):
    """Convert atom symbols and positions to the PDB format.

    File format definitions:
        CRYST1: https://wwpdb.org/documentation/file-format-content/format33/sect8.html#CRYST1

        ATOM: https://wwpdb.org/documentation/file-format-content/format33/sect9.html#ATOM

    Args:
        atom (list): Atom symbols.
        X (ndarray): Atom positions.

    Keyword Args:
        a (float): Cell size.

    Returns:
        str: PDB file format.
    """
    # Convert Bohr to Angstrom
    X = bohr2ang(X)
    if a is not None:
        a = bohr2ang(a)

    # pdb files have specific numbers of characters for every data with changing justification
    # Write everything explicitly down to not lose track of line lengths
    pdb = ''
    # Create data for a cuboidal cell
    if a is not None:
        pdb += 'CRYST1'          # 1-6 "CRYST1"
        pdb += f'{a[0]:>9,.3f}'  # 7-15 a
        pdb += f'{a[1]:>9,.3f}'  # 16-24 b
        pdb += f'{a[2]:>9,.3f}'  # 25-33 c
        pdb += f'{90:>7,.2f}'    # 34-40 alpha
        pdb += f'{90:>7,.2f}'    # 41-47 beta
        pdb += f'{90:>7,.2f}'    # 48-54 gamma
        pdb += ' '
        pdb += 'P 1        '     # 56-66 Space group
        # pdb += '   1'          # 67-70 Z value
        pdb += '\n'

    # Create molecule data
    pdb += 'MODEL 1'
    for ia in range(len(atom)):
        pdb += '\nATOM  '            # 1-6 "ATOM"
        pdb += f'{ia + 1:>5}'        # 7-11 Atom serial number
        pdb += ' '
        pdb += f'{atom[ia]:>4}'      # 13-16 Atom name
        pdb += ' '                   # 17 Alternate location indicator
        pdb += 'MOL'                 # 18-20 Residue name
        pdb += ' '
        pdb += ' '                   # 22 Chain identifier
        pdb += '   1'                # 23-26 Residue sequence number
        pdb += ' '                   # 27 Code for insertions of residues
        pdb += '   '
        pdb += f'{X[ia, 0]:>8,.3f}'  # 31-38 X orthogonal coordinate
        pdb += f'{X[ia, 1]:>8,.3f}'  # 39-46 Y orthogonal coordinate
        pdb += f'{X[ia, 2]:>8,.3f}'  # 47-54 Z orthogonal coordinate
        pdb += f'{1:>6,.2f}'         # 55-60 Occupancy
        pdb += f'{0:>6,.2f}'         # 61-66 Temperature factor
        pdb += '          '
        pdb += f'{atom[ia]:>2}'      # 77-78 Element symbol
        # pdb += '  '                # 79-80 Charge
    return f'{pdb}\nENDMDL\n'



#!/usr/bin/env python3
"""JSON file handling."""
import base64
import copy
import json

import numpy as np


def read_json(filename):
    """Load objects from a JSON file.

    Args:
        filename (str): JSON input file path/name.

    Returns:
        Class object.
    """
    import eminus

    def set_attrs(obj, dct):
        """Set attributes of an object using a given dictionary."""
        for attr in dct:
            if attr == 'log':
                continue
            setattr(obj, attr, copy.deepcopy(dct[attr]))
        return obj

    def custom_object_hook(dct):
        """Custom JSON object hook to create eminus classes after deserialization."""
        # ndarrays are base64 encoded, decode and recreate
        if isinstance(dct, dict) and '__ndarray__' in dct:
            data = base64.b64decode(dct['__ndarray__'])
            return np.frombuffer(data, dct['dtype']).reshape(dct['shape'])

        # Create simple eminus objects and set all attributes afterwards
        # Explicitly call objects with verbosity since the logger is created at instantiation

        # Atoms objects
        if isinstance(dct, dict) and '_atom' in dct:
            atoms = eminus.Atoms(dct['_atom'], dct['_X'], verbose=dct['_verbose'])
            atoms._set_operators()
            atoms = set_attrs(atoms, dct)
            # The tuple type is not preserved when serializing, manually cast the only important one
            if isinstance(atoms._active, list):
                atoms._active = tuple(atoms._active)
            return atoms
        # SCF objects
        if isinstance(dct, dict) and '_atoms' in dct:
            scf = eminus.SCF(dct['_atoms'], verbose=dct['_verbose'])
            return set_attrs(scf, dct)
        # Energy objects
        if isinstance(dct, dict) and 'Ekin' in dct:
            energies = eminus.energies.Energy()
            return set_attrs(energies, dct)
        # GTH objects
        if isinstance(dct, dict) and 'NbetaNL' in dct:
            gth = eminus.gth.GTH()
            return set_attrs(gth, dct)
        # Occupations objects
        if isinstance(dct, dict) and '_Nelec' in dct:
            occ = eminus.occupations.Occupations()
            return set_attrs(occ, dct)
        return dct

    if not filename.endswith('.json'):
        filename += '.json'

    with open(filename, 'r') as fh:
        return json.load(fh, object_hook=custom_object_hook)


def write_json(obj, filename):
    """Save objects in a JSON file.

    Args:
        obj: Class object.
        filename (str): JSON output file path/name.
    """
    import eminus

    class CustomEncoder(json.JSONEncoder):
        """Custom JSON encoder class to serialize eminus classes."""
        def default(self, obj):
            # ndarrays are not json serializable, encode them as base64 to save them
            if isinstance(obj, np.ndarray):
                data = base64.b64encode(obj.copy(order='C')).decode('utf-8')
                return {'__ndarray__': data, 'dtype': str(obj.dtype), 'shape': obj.shape}

            # If obj is an eminus class dump them as a dictionary
            if isinstance(obj, (eminus.Atoms, eminus.SCF, eminus.energies.Energy, eminus.gth.GTH,
                                eminus.occupations.Occupations)):
                # Only dumping the dict would result in a string, so do one dump and one load
                data = json.dumps(obj.__dict__, cls=CustomEncoder)
                return dict(json.loads(data))
            # The logger class is not serializable, just ignore it
            if isinstance(obj, eminus.logger.CustomLogger):
                return None
            return json.JSONEncoder.default(self, obj)

    if not filename.endswith('.json'):
        filename += '.json'

    with open(filename, 'w') as fp:
        json.dump(obj, fp, cls=CustomEncoder)



#!/usr/bin/env python3
"""Package version number and version info function."""
import importlib
import platform
import sys

__version__ = '2.6.0'
#: eminus ASCII logo.
LOGO = (' ___ _____ _ ___ _ _ ___ \n'
        '| -_|     | |   | | |_ -|\n'
        '|___|_|_|_|_|_|_|___|___|\n')


def info():
    """Print version numbers and availability of packages."""
    dependencies = ('numpy', 'scipy')
    extras = ('torch', 'pyscf', 'plotly', 'nglview')
    dev = ('notebook', 'pylibxc', 'pytest', 'sphinx', 'furo')

    print(LOGO)
    print('--- Platform infos ---'
          f'\nPlatform : {platform.system()} {platform.machine()}'
          f'\nRelease  : {platform.release()} {platform.version()}'
          '\n\n--- Version infos ---'
          f'\npython   : {sys.version.split()[0]}'
          f'\neminus   : {__version__}')
    for pkg in dependencies + extras + dev:
        try:
            module = importlib.import_module(pkg)
            try:
                print(f'{pkg:<9}: {module.__version__}')
            except AttributeError:
                # pylibxc does not use the standard version identifier
                print(f'{pkg:<9}: {module.version.__version__}')
        except ModuleNotFoundError:
            if pkg in dependencies:
                print(f'{pkg:<9}: Dependency not installed')
            elif pkg in extras:
                print(f'{pkg:<9}: Extra not installed')


if __name__ == '__main__':
    info()



#!/usr/bin/env python3
"""File input and output functionalities."""
from .cube import read_cube, write_cube
from .gth import read_gth
from .json import read_json, write_json
from .pdb import create_pdb_str, write_pdb
from .xyz import read_xyz, write_xyz

__all__ = ['create_pdb_str', 'read', 'read_cube', 'read_gth', 'read_json', 'read_xyz', 'write',
           'write_cube', 'write_json', 'write_pdb', 'write_xyz']


def read(*args, **kwargs):
    """Unified file reader function."""
    if args[0].endswith('.json'):
        return read_json(*args, **kwargs)
    if args[0].endswith('.xyz'):
        return read_xyz(*args, **kwargs)
    if args[0].endswith('.cube'):
        return read_cube(*args, **kwargs)
    NotImplementedError('File ending not recognized.')
    return None


def write(*args, **kwargs):
    """Unified file writer function."""
    if args[1].endswith('.json'):
        return write_json(*args, **kwargs)
    if args[1].endswith('.xyz'):
        return write_xyz(*args, **kwargs)
    if args[1].endswith('.cube'):
        return write_cube(*args, **kwargs)
    if args[1].endswith('.pdb'):
        return write_pdb(*args, **kwargs)
    NotImplementedError('File ending not recognized.')
    return None



#!/usr/bin/env python3
"""Chachiyo GGA correlation.

Reference: Comput. Theor. Chem. 1172, 112669.
"""
import numpy as np
from numpy.linalg import norm

from .lda_c_chachiyo_mod import chachiyo_scaling_mod as weight_function


def gga_c_chachiyo(n, dn_spin=None, **kwargs):
    """Chachiyo parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label GGA_C_CHACHIYO and ID 309 in Libxc.

    Reference: Comput. Theor. Chem. 1172, 112669.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: Chachiyo correlation energy density, potential and vsigma.
    """
    h = 0.06672632  # 0.5 * 0.00847 * 16 * (3 / np.pi)**(1 / 3)

    # ### Start lda_c_chachiyo_mod ### #
    a = -0.01554535  # (np.log(2) - 1) / (2 * np.pi**2)
    b = 20.4562557

    rs = (3 / (4 * np.pi * n))**(1 / 3)
    rs2 = rs**2
    ecinner = 1 + b / rs + b / rs2

    ec = a * np.log(ecinner)
    # ### End lda_c_chachiyo_mod ### #

    norm_dn = norm(dn_spin[0], axis=1)
    t = (np.pi / 3)**(1 / 6) / 4 * norm_dn / n**(7 / 6)
    t2 = t**2
    gec = 1 + t2
    expgec = gec**(h / ec)

    # ### Start lda_c_chachiyo_mod ### #
    vc = ec + a * b * (2 + rs) / (3 * (b + b * rs + rs2))
    # ### End lda_c_chachiyo_mod ### #

    term1 = h * (1 - 1 / gec)
    term2 = h * np.log(gec) * (1 - vc / ec)
    gvc = (vc - 7 / 3 * term1 + term2) * expgec

    vsigmac = n * expgec * term1 / norm_dn**2
    return ec * expgec, np.array([gvc]), np.array([vsigmac])


def gga_c_chachiyo_spin(n, zeta, dn_spin=None, **kwargs):
    """Chachiyo parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label GGA_C_CHACHIYO and ID 309 in Libxc.

    Reference: Comput. Theor. Chem. 1172, 112669.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: Chachiyo correlation energy density, potential and vsigma.
    """
    h = 0.06672632  # 0.5 * 0.00847 * 16 * (3 / np.pi)**(1 / 3)

    # ### Start lda_c_chachiyo_spin_mod ### #
    a0 = -0.01554535   # (np.log(2) - 1) / (2 * np.pi**2)
    a1 = -0.007772675  # (np.log(2) - 1) / (4 * np.pi**2)
    b0 = 20.4562557
    b1 = 27.4203609

    rs = (3 / (4 * np.pi * n))**(1 / 3)
    rs2 = rs**2

    fzeta, dfdzeta = weight_function(zeta)

    ec0inner = 1 + b0 / rs + b0 / rs2
    ec1inner = 1 + b1 / rs + b1 / rs2
    ec0 = a0 * np.log(ec0inner)
    ec1 = a1 * np.log(ec1inner)

    ec = ec0 + (ec1 - ec0) * fzeta
    # ### End lda_c_chachiyo_spin_mod ### #

    norm_dn = norm(dn_spin[0] + dn_spin[1], axis=1)
    t = (np.pi / 3)**(1 / 6) / 4 * norm_dn / n**(7 / 6)
    t2 = t**2
    gec = 1 + t2
    expgec = gec**(h / ec)

    # ### Start lda_c_chachiyo_spin_mod ### #
    factor = -1 / rs2 - 2 / rs**3
    dec0drs = a0 / ec0inner * b0 * factor
    dec1drs = a1 / ec1inner * b1 * factor
    decdrs = dec0drs + (dec1drs - dec0drs) * fzeta
    # prefactor = ec - rs / 3 * decdrs
    decdf = (ec1 - ec0) * dfdzeta

    # vcup = prefactor + decdf * (1 - zeta)
    # vcdw = prefactor - decdf * (1 + zeta)
    # ### End lda_c_chachiyo_spin_mod ### #

    dn2 = norm(dn_spin[0], axis=1)**2 + \
        2 * np.sum(dn_spin[0] * dn_spin[1], axis=1) + \
        norm(dn_spin[1], axis=1)**2
    ht2divgecdn2 = (1 - 1 / gec) * h / norm_dn**2
    term1 = -7 / 3 * ht2divgecdn2 * dn2
    term2 = 1 - h * np.log(gec) / ec
    prefactor = -decdrs * rs / 3
    gvc = ec + term1 + term2 * prefactor
    gvcup = gvc + term2 * decdf * (1 - zeta)
    gvcdw = gvc - term2 * decdf * (1 + zeta)

    vsigma = n * expgec * 2 * ht2divgecdn2
    vsigmac = np.array([0.5 * vsigma, vsigma, 0.5 * vsigma])
    return ec * expgec, np.array([gvcup, gvcdw]) * expgec, vsigmac



#!/usr/bin/env python3
"""Atoms class definition."""
import numbers

import numpy as np
from scipy.fft import next_fast_len
from scipy.linalg import det, eigh, inv, norm

from . import config, operators
from .logger import create_logger, get_level, log
from .occupations import Occupations
from .tools import center_of_mass, cutoff2gridspacing, inertia_tensor
from .utils import atom2charge, molecule2list


class Atoms:
    """Atoms object that holds all system and cell parameters.

    Args:
        atom (str | list | tuple): Atom symbols.

            A string can be given, e.g., with :code:`CH4` that will be parsed to
            :code:`['C', 'H', 'H', 'H', 'H']`. When calculating atoms one can directly provide the
            charge, e.g., with :code:`Li-q3`.
        X (list | tuple | ndarray): Atom positions.

    Keyword Args:
        ecut (float | None): Cut-off energy.

            Defaults to 30 Eh (ca. 816 eV).
        a (float | list | tuple | ndarray): Cell size or vacuum size.

            Floats will create a cubic unit cell. Defaults to 20 a0 (ca. 10.5 A).
        spin (int | None): Number of unpaired electrons.

            This is the difference between the number of up and down electrons.
        charge (int): Charge of the system.
        unrestricted (bool | None): Handling of spin.

            :code:`False` for restricted, :code:`True` for unrestricted, and :code:`None` for
            automatic detection.
        center (bool | str): Center the system inside the cell.

            Aligns the geometric center of mass with the center of the call and rotates the system,
            such that its geometric moment of inertia aligns with the coordinate axes. Can be one of
            bool, 'shift', and 'rotate'.
        verbose (int | str | None): Level of output.

            Can be one of 'critical', 'error', 'warning', 'info' (default), or 'debug'. An integer
            value can be used as well, where larger numbers mean more output, starting from 0.
            None will use the global logger verbosity value.
    """
    def __init__(self, atom, X, ecut=30, a=20, spin=None, charge=0, unrestricted=None, center=False,
                 verbose=None):
        """Initialize the Atoms object."""
        # Set the input parameters (the ordering is important)
        self.log = create_logger(self)    #: Logger object.
        self.verbose = verbose            #: Verbosity level.
        self.occ = Occupations()          #: Occupations object.
        self.atom = atom                  #: Atom symbols.
        self.X = X                        #: Atom positions.
        self.a = a                        #: Cell/Vacuum size.
        self.ecut = ecut                  #: Cut-off energy.
        self.center = center              #: Enables centering the system in the cell.
        self.charge = charge              #: System charge.
        self.spin = spin                  #: Number of unpaired electrons.
        self.unrestricted = unrestricted  #: Enables unrestricted spin handling.

        # Initialize other attributes
        self.occ.fill()                   #: Fill states from the given input.
        self.is_built = False             #: Determines the Atoms object build status.

    # ### Class properties ###

    @property
    def atom(self):
        """Atom symbols."""
        return self._atom

    @atom.setter
    def atom(self, value):
        # Quick option to set the charge for single atoms
        if isinstance(value, str) and '-q' in value:
            atom, Z = value.split('-q')
            self._atom = [atom]
            self._Natoms = 1
            self.Z = int(Z)
        else:
            # If a string, i.e., chemical formula is given convert it to a list of chemical symbols
            if isinstance(value, str):
                self._atom = molecule2list(value)
            else:
                self._atom = value
            # Get the number of atoms and determine the charges
            self._Natoms = len(self._atom)
            self.Z = None

    @property
    def X(self):
        """Atom positions."""
        return self._X

    @X.setter
    def X(self, value):
        # We need atom positions as a two-dimensional array
        self._X = np.atleast_2d(value)
        if self.Natoms != len(self._X):
            raise ValueError(f'Mismatch between number of atoms ({self.Natoms}) and number of '
                             f'coordinates ({len(self._X)}).')
        # The structure factor changes when changing X
        self.is_built = False

    @property
    def ecut(self):
        """Cut-off energy."""
        return self._ecut

    @ecut.setter
    def ecut(self, value):
        self._ecut = value
        # Caclulate the sampling from the cut-off energy
        s = np.int_(self.a / cutoff2gridspacing(value))
        # Multiply by two and add one to match PWDFT.jl
        s = 2 * s + 1
        # Calculate a fast length to optimize the FFT calculations
        self.s = [next_fast_len(i) for i in s]
        # The cell discretization changes when changing s or ecut
        self.is_built = False

    @property
    def a(self):
        """Cell/Vacuum size."""
        return self._a

    @a.setter
    def a(self, value):
        # Build a cubic cell if a number is given
        if isinstance(value, numbers.Real):
            self._a = value * np.ones(3)
        else:
            self._a = np.asarray(value)
        # Build a cuboidal unit cell and calculate its volume
        self._R = self._a * np.eye(3)
        self._Omega = abs(det(self._R))
        # The cell changes when changing a
        self.is_built = False

    @property
    def spin(self):
        """Number of unpaired electrons."""
        return self.occ.spin

    @spin.setter
    def spin(self, value):
        self.occ.spin = value

    @property
    def charge(self):
        """System charge."""
        return self.occ.charge

    @charge.setter
    def charge(self, value):
        self.occ.charge = value

    @property
    def unrestricted(self):
        """Enables unrestricted spin handling."""
        return self.occ.Nspin == 2

    @unrestricted.setter
    def unrestricted(self, value):
        if value is None:
            self.occ.Nspin = value
        else:
            self.occ.Nspin = value + 1

    @property
    def center(self):
        """Enables centering the system in the cell."""
        return self._center

    @center.setter
    def center(self, value):
        if isinstance(value, str):
            self._center = value.lower()
            if self._center not in ('rotate', 'shift', 'recentered'):
                log.error(f'{self._center} is not a recognized center method.')
        else:
            self._center = value
        # Do nothing when recentering
        if self._center == 'recentered':
            return
        # Center system such that the geometric inertia tensor will be diagonal
        # Rotate before shifting!
        if self._center is True or self._center == 'rotate':
            I = inertia_tensor(self.X)
            _, eigvecs = eigh(I)
            self.X = (inv(eigvecs) @ self.X.T).T
        # Shift system such that its geometric center of mass is in the center of the cell
        if self._center is True or self._center == 'shift':
            com = center_of_mass(self.X)
            self.X = self.X - (com - self.a / 2)
        # The structure factor changes when changing X
        self.is_built = False

    @property
    def verbose(self):
        """Verbosity level."""
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        # If no verbosity is given use the global verbosity level
        if value is None:
            value = log.verbose
        self._verbose = get_level(value)
        self.log.verbose = self._verbose

    # ### Class properties with a setter outside of the init method ###

    @property
    def f(self):
        """Occupation numbers per state."""
        return self.occ.f

    @f.setter
    def f(self, value):
        # Pass through to the Occupations object
        self.occ.f = value

    @property
    def s(self):
        """Real-space sampling of the cell."""
        return self._s

    @s.setter
    def s(self, value):
        # Choose the same sampling for every direction if an integer is given
        if isinstance(value, numbers.Integral):
            value = value * np.ones(3, dtype=int)
        self._s = np.asarray(value)
        self._Ns = int(np.prod(self._s))
        # The cell discretization changes when changing s
        self.is_built = False

    @property
    def Z(self):
        """Valence charge per atom."""
        return self._Z

    @Z.setter
    def Z(self, value):
        # Assume same charges for all atoms if an integer is given
        if isinstance(value, numbers.Integral):
            value = value * np.ones(self.Natoms, dtype=int)
        elif isinstance(value, dict):
            value = [value[ia] for ia in self.atom]
        # Get the valence charges from the GTH files
        elif value is None or isinstance(value, str):
            value = atom2charge(self.atom, value)
        self._Z = np.asarray(value)
        if self.Natoms != len(self._Z):
            raise ValueError(f'Mismatch between number of atoms ({self.Natoms}) and number of '
                             f'charges ({len(self._Z)}).')
        # Get the number of calculated electrons and pass it to occ
        self.occ.Nelec = np.sum(self._Z) - self.charge

    # ### Read-only properties ###

    @property
    def Natoms(self):
        """Number of atoms."""
        return self._Natoms

    @property
    def Ns(self):
        """Number of real-space grid points."""
        return self._Ns

    @property
    def Omega(self):
        """Unit cell volume."""
        return self._Omega

    @property
    def r(self):
        """Real-space sampling points."""
        return self._r

    @property
    def R(self):
        """Unit cell vectors."""
        return self._R

    @property
    def active(self):
        """Mask for active G-vectors."""
        return self._active

    @property
    def G(self):
        """G-vectors."""
        return self._G

    @property
    def G2(self):
        """Squared magnitudes of G-vectors."""
        return self._G2

    @property
    def G2c(self):
        """Truncated squared magnitudes of G-vectors."""
        return self._G2c

    @property
    def Sf(self):
        """Structure factor per atom."""
        return self._Sf

    @property
    def dV(self):
        """Volume element to multiply when integrating field properties."""
        return self.Omega / self._Ns

    @property
    def _atoms(self):
        """Return the Atoms object itself."""
        # This way we can access the object from Atoms and SCF classes with the same code
        return self

    # ### Class methods ###

    def build(self):
        """Build all parameters of the Atoms object."""
        self._set_operators()
        self._sample_unit_cell()
        self.occ.fill()
        self.is_built = True
        return self

    kernel = build

    def recenter(self, center=None):
        """Recenter the system inside the cell.

        Keyword Args:
            center (float | list | tuple | ndarray | None): Point to center the system around.
        """
        com = center_of_mass(self.X)
        if center is None:
            self.X = self.X - (com - self.a / 2)
        else:
            center = np.asarray(center)
            self.X = self.X - (com - center)
        # Recalculate the structure factor since it depends on the atom positions
        self._Sf = np.exp(1j * self.G @ self.X.T).T
        self._center = 'recentered'
        return self

    def clear(self):
        """Initialize or clear parameters that will be built out of the inputs."""
        self._r = None          # Sample points in cell
        self._G = None          # G-vectors
        self._G2 = None         # Squared magnitudes of G-vectors
        self._active = None     # Mask for active G-vectors
        self._G2c = None        # Truncated squared magnitudes of G-vectors
        self._Sf = None         # Structure factor
        self.is_built = False   # Flag to determine if the object was built or not
        return self

    def _get_index_matrices(self):
        """Build index matrices M and N to build the real and reciprocal space samplings.

        The matrices are using C ordering (the last index is the fastest).

        Returns:
            tuple[ndarray, ndarray]: Index matrices.
        """
        # Build index matrix M
        ms = np.arange(self._Ns)
        m1 = np.floor(ms / (self.s[2] * self.s[1])) % self.s[0]
        m2 = np.floor(ms / self.s[2]) % self.s[1]
        m3 = ms % self.s[2]
        M = np.column_stack((m1, m2, m3))
        # Build index matrix N
        n1 = m1 - (m1 > self.s[0] / 2) * self.s[0]
        n2 = m2 - (m2 > self.s[1] / 2) * self.s[1]
        n3 = m3 - (m3 > self.s[2] / 2) * self.s[2]
        N = np.column_stack((n1, n2, n3))
        return M, N

    def _sample_unit_cell(self):
        """Build the real-space sampling and all G-vector parameters."""
        # Calculate index matrices
        M, N = self._get_index_matrices()
        # Build the real-space sampling
        self._r = M @ inv(np.diag(self.s)) @ self.R.T
        # Build G-vectors
        self._G = 2 * np.pi * N @ inv(self.R)
        # Calculate squared magnitudes of G-vectors
        self._G2 = norm(self.G, axis=1)**2
        # Calculate the G2 restriction
        self._active = np.nonzero(2 * self.ecut >= self.G2)
        self._G2c = self.G2[self._active]
        # Calculate the structure factor per atom
        self._Sf = np.exp(1j * self.G @ self.X.T).T

    def _base_operator(*args, **kwargs):
        """See :mod:`~eminus.operators`."""
        # Base operator method to show an error for unbuilt Atoms objects
        log.error('Build the Atoms object with "atoms.build()" to call this operator.')

    def _set_operators(self):
        """Set operators of an Atoms class instance at runtime."""
        for op in ('O', 'L', 'Linv', 'K', 'T'):
            setattr(type(self), op, getattr(operators, op))
        fft_operators = ('I', 'J', 'Idag', 'Jdag')
        # Use the Torch operators if desired, or the default ones otherwise
        if config.use_torch:
            from .extras import torch
            for op in fft_operators:
                setattr(type(self), op, getattr(torch, op))
        else:
            for op in fft_operators:
                setattr(type(self), op, getattr(operators, op))

    O = _base_operator
    L = _base_operator
    Linv = _base_operator
    K = _base_operator
    T = _base_operator
    I = _base_operator
    J = _base_operator
    Idag = _base_operator
    Jdag = _base_operator

    def __repr__(self):
        """Print the parameters stored in the Atoms object."""
        out = 'Atom  Valence  Position'
        for i in range(self.Natoms):
            out += f'\n{self.atom[i]:>3}   {self.Z[i]:>6}   ' \
                   f'{self.X[i, 0]:10.5f}  {self.X[i, 1]:10.5f}  {self.X[i, 2]:10.5f}'
        return out



#!/usr/bin/env python3
"""Collection of constants and unit conversion functions.

For more about atomic units, see: https://en.wikipedia.org/wiki/Hartree_atomic_units
"""
#: Hartree in electronvolt.
electronvolt = eV = 27.211386245988  # https://en.wikipedia.org/wiki/Hartree
#: Hartree in kcal per mol.
kcalmol = 627.5094740631  # https://en.wikipedia.org/wiki/Hartree
#: Bohr radius in Angstrom.
Angstrom = A = 0.529177210903  # https://en.wikipedia.org/wiki/Bohr_radius
#: Elementary charge times bohr radius in Debye.
Debye = D = 2.541746473  # https://en.wikipedia.org/wiki/Hartree_atomic_units


def ha2ev(E):
    """Convert Hartree to electronvolt.

    Args:
        E (float | ndarray): Energy in Hartree.

    Returns:
        float | ndarray: Energy in electronvolt.
    """
    return E * electronvolt


def ev2ha(E):
    """Convert electronvolt to Hartree.

    Args:
        E (float | ndarray): Energy in electronvolt.

    Returns:
        float | ndarray: Energy in Hartree.
    """
    return E / electronvolt


def ha2kcalmol(E):
    """Convert Hartree to kcal/mol.

    Args:
        E (float | ndarray): Energy in Hartree.

    Returns:
        float | ndarray: Energy in kcal/mol.
    """
    return E * kcalmol


def kcalmol2ha(E):
    """Convert kcal/mol to Hartree.

    Args:
        E (float | ndarray): Energy in kcal/mol.

    Returns:
        float | ndarray: Energy in Hartree.
    """
    return E / kcalmol


def ha2ry(E):
    """Convert Hartree to Rydberg.

    Args:
        E (float | ndarray): Energy in Hartree.

    Returns:
        float | ndarray: Energy in Rydberg.
    """
    return 2 * E


def ry2ha(E):
    """Convert Rydberg to Hartree.

    Args:
        E (float | ndarray): Energy in Rydberg.

    Returns:
        float | ndarray: Energy in Hartree.
    """
    return E / 2


def ang2bohr(r):
    """Convert Angstrom to Bohr.

    Args:
        r (float | ndarray): Length in Angstrom.

    Returns:
        float | ndarray: Length in Bohr.
    """
    return r / Angstrom


def bohr2ang(r):
    """Convert Bohr to Angstrom.

    Args:
        r (float | ndarray): Length in Bohr.

    Returns:
        float | ndarray: Length in Angstrom.
    """
    return r * Angstrom


def ebohr2d(p):
    """Convert e * Bohr to Debye.

    Args:
        p (float | ndarray): Electric dipole moment in e * Bohr.

    Returns:
        float | ndarray: Electric dipole moment in Debye.
    """
    return p * Debye


def d2ebohr(p):
    """Convert Debye to e * Bohr.

    Args:
        p (float | ndarray): Electric dipole moment in Debye.

    Returns:
        float | ndarray: Electric dipole moment in e * Bohr.
    """
    return p / Debye



#!/usr/bin/env python3
"""Perdew-Burke-Ernzerhof GGA correlation.

Reference: Phys. Rev. Lett. 78, 1396.
"""
import numpy as np
from numpy.linalg import norm

from .lda_c_pw_mod import lda_c_pw_mod, lda_c_pw_mod_spin


def gga_c_pbe(n, beta=0.06672455060314922, dn_spin=None, **kwargs):
    """Perdew-Burke-Ernzerhof parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label GGA_C_PBE and ID 130 in Libxc.

    Reference: Phys. Rev. Lett. 78, 1396.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        beta (float): Functional parameter.
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: PBE correlation energy density, potential, and vsigma.
    """
    gamma = (1 - np.log(2)) / np.pi**2

    pi34 = (3 / (4 * np.pi))**(1 / 3)
    rs = pi34 * n**(-1 / 3)
    norm_dn = norm(dn_spin[0], axis=1)
    ec, vc, _ = lda_c_pw_mod(n, **kwargs)

    kf = (9 / 4 * np.pi)**(1 / 3) / rs
    ks = np.sqrt(4 * kf / np.pi)
    divt = 2 * ks * n
    t = norm_dn / divt
    expec = np.exp(-ec / gamma)
    A = beta / (gamma * (expec - 1))
    t2 = t**2
    At2 = A * t2
    A2t4 = At2**2
    divsum = 1 + At2 + A2t4
    div = (1 + At2) / divsum
    nolog = 1 + beta / gamma * t2 * div
    gec = gamma * np.log(nolog)

    factor = A2t4 * (2 + At2) / divsum**2
    dgec = beta * t2 / nolog * (-7 / 3 * div - factor * (A * expec * (vc - ec) / beta - 7 / 3))
    gvc = gec + dgec

    vsigmac = beta / (divt * ks) * (div - factor) / nolog
    return ec + gec, np.array([vc + gvc]), np.array([0.5 * vsigmac])


def gga_c_pbe_spin(n, zeta, beta=0.06672455060314922, dn_spin=None, **kwargs):
    """Perdew-Burke-Ernzerhof parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label GGA_C_PBE and ID 130 in Libxc.

    Reference: Phys. Rev. Lett. 78, 1396.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        beta (float): Functional parameter.
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: PBE correlation energy density, potential, and vsigma.
    """
    gamma = (1 - np.log(2)) / np.pi**2

    pi34 = (3 / (4 * np.pi))**(1 / 3)
    rs = pi34 * n**(-1 / 3)
    norm_dn = norm(dn_spin[0] + dn_spin[1], axis=1)
    ec, vc, _ = lda_c_pw_mod_spin(n, zeta, **kwargs)
    vcup, vcdw = vc

    kf = (9 / 4 * np.pi)**(1 / 3) / rs
    ks = np.sqrt(4 * kf / np.pi)
    phi = ((1 + zeta)**(2 / 3) + (1 - zeta)**(2 / 3)) / 2
    phi2 = phi**2
    phi3 = phi2 * phi
    t = norm_dn / (2 * phi * ks * n)
    expec = np.exp(-ec / (gamma * phi3))
    A = beta / (gamma * (expec - 1))
    t2 = t**2
    At2 = A * t2
    A2t4 = At2**2
    divsum = 1 + At2 + A2t4
    div = (1 + At2) / divsum
    nolog = 1 + beta / gamma * t2 * div
    gec = gamma * phi3 * np.log(nolog)

    # Handle divisions by zero
    with np.errstate(divide='ignore', invalid='ignore'):
        dfz = ((1 + zeta)**(-1 / 3) - (1 - zeta)**(-1 / 3)) / 3
    dfz = np.nan_to_num(dfz, nan=0, posinf=0, neginf=0)
    factor = A2t4 * (2 + At2) / divsum**2
    bfpre = expec / phi3
    bfup = bfpre * (vcup - ec)
    bfdw = bfpre * (vcdw - ec)
    dgecpre = beta * t2 * phi3 / nolog
    dgecup = dgecpre * (-7 / 3 * div - factor * (A * bfup / beta - 7 / 3))
    dgecdw = dgecpre * (-7 / 3 * div - factor * (A * bfdw / beta - 7 / 3))
    dgeczpre = (3 * gec / phi - beta * t2 * phi2 / nolog * (
                2 * div - factor * (3 * A * expec * ec / phi3 / beta + 2))) * dfz
    dgeczup = dgeczpre * (1 - zeta)
    dgeczdw = -dgeczpre * (1 + zeta)
    gvcup = gec + dgecup + dgeczup
    gvcdw = gec + dgecdw + dgeczdw

    vsigma = beta * phi / (2 * ks * ks * n) * (div - factor) / nolog
    vsigmac = np.array([0.5 * vsigma, vsigma, 0.5 * vsigma])
    return ec + gec, np.array([vcup + gvcup, vcdw + gvcdw]), vsigmac



#!/usr/bin/env python3
"""Linear algebra calculation utilities."""
import functools
import re

import numpy as np
from scipy.linalg import norm

from .io import read_gth


def dotprod(a, b):
    """Efficiently calculate the expression a * b.

    Add an extra check to make sure the result is never zero since this function is used as a
    denominator in minimizers.

    Args:
        a (ndarray): Array of vectors.
        b (ndarray): Array of vectors.

    Returns:
        float: The expressions result
    """
    eps = 1e-15  # 2.22e-16 is the range of float64 machine precision
    # The dot product of complex vectors looks like the expression below, but this is slow
    # res = np.real(np.trace(a.conj().T @ b))
    # We can calculate the trace faster by taking the sum of the Hadamard product
    res = np.sum(a.conj() * b)
    if abs(res) < eps:
        return eps
    return np.real(res)


def Ylm_real(l, m, G):  # noqa: C901
    """Calculate real spherical harmonics from cartesian coordinates.

    Reference: https://scipython.com/blog/visualizing-the-real-forms-of-the-spherical-harmonics

    Args:
        l (int): Angular momentum number.
        m (int): Magnetic quantum number.
        G (ndarray): Reciprocal lattice vector or array of lattice vectors.

    Returns:
        ndarray: Real spherical harmonics.
    """
    eps = 1e-9
    # Account for single vectors
    G = np.atleast_2d(G)

    # No need to calculate more for l=0
    if l == 0:
        return 0.5 * np.sqrt(1 / np.pi) * np.ones(len(G))

    # cos(theta)=Gz/|G|
    Gm = norm(G, axis=1)
    with np.errstate(divide='ignore', invalid='ignore'):
        cos_theta = G[:, 2] / Gm
    # Account for small magnitudes, if norm(G) < eps: cos_theta=0
    cos_theta[Gm < eps] = 0

    # Vectorized version of sin(theta)=sqrt(max(0, 1-cos_theta^2))
    sin_theta = np.sqrt(np.amax((np.zeros_like(cos_theta), 1 - cos_theta**2), axis=0))

    # phi=arctan(Gy/Gx)
    phi = np.arctan2(G[:, 1], G[:, 0])
    # If Gx=0: phi=pi/2*sign(Gy)
    phi_idx = np.abs(G[:, 0]) < eps
    phi[phi_idx] = np.pi / 2 * np.sign(G[phi_idx, 1])

    if l == 1:
        if m == -1:   # py
            return 0.5 * np.sqrt(3 / np.pi) * sin_theta * np.sin(phi)
        if m == 0:  # pz
            return 0.5 * np.sqrt(3 / np.pi) * cos_theta
        if m == 1:  # px
            return 0.5 * np.sqrt(3 / np.pi) * sin_theta * np.cos(phi)
    elif l == 2:
        if m == -2:    # dxy
            return np.sqrt(15 / 16 / np.pi) * sin_theta**2 * np.sin(2 * phi)
        if m == -1:  # dyz
            return np.sqrt(15 / 4 / np.pi) * cos_theta * sin_theta * np.sin(phi)
        if m == 0:   # dz2
            return 0.25 * np.sqrt(5 / np.pi) * (3 * cos_theta**2 - 1)
        if m == 1:   # dxz
            return np.sqrt(15 / 4 / np.pi) * cos_theta * sin_theta * np.cos(phi)
        if m == 2:   # dx2-y2
            return np.sqrt(15 / 16 / np.pi) * sin_theta**2 * np.cos(2 * phi)
    elif l == 3:
        if m == -3:
            return 0.25 * np.sqrt(35 / 2 / np.pi) * sin_theta**3 * np.sin(3 * phi)
        if m == -2:
            return 0.25 * np.sqrt(105 / np.pi) * sin_theta**2 * cos_theta * np.sin(2 * phi)
        if m == -1:
            return 0.25 * np.sqrt(21 / 2 / np.pi) * sin_theta * (5 * cos_theta**2 - 1) * np.sin(phi)
        if m == 0:
            return 0.25 * np.sqrt(7 / np.pi) * (5 * cos_theta**3 - 3 * cos_theta)
        if m == 1:
            return 0.25 * np.sqrt(21 / 2 / np.pi) * sin_theta * (5 * cos_theta**2 - 1) * np.cos(phi)
        if m == 2:
            return 0.25 * np.sqrt(105 / np.pi) * sin_theta**2 * cos_theta * np.cos(2 * phi)
        if m == 3:
            return 0.25 * np.sqrt(35 / 2 / np.pi) * sin_theta**3 * np.cos(3 * phi)

    ValueError(f'No definition found for Ylm({l}, {m}).')
    return None


def handle_spin_gracefully(func):
    """Handle spin calculating the function for each channel separately.

    This can only be applied if the only spin-dependent indexing is the wave function W.

    Implementing the explicit handling of spin adds an extra layer of complexity where one has to
    loop over the spin states in many places. We can hide this complexity using this decorator while
    still supporting many use-cases, e.g., the operators previously act on arrays containing wave
    functions of all states and of one state only. This decorator maintains this functionality and
    adds the option to act on arrays containing wave functions of all spins and all states as well.

    Args:
        func (Callable): Function that acts on spin-states.

    Returns:
        Callable: Decorator.
    """
    @functools.wraps(func)
    def decorator(obj, W, *args, **kwargs):
        if W.ndim == 3:
            return np.asarray([func(obj, Wspin, *args, **kwargs) for Wspin in W])
        return func(obj, W, *args, **kwargs)
    return decorator


def pseudo_uniform(size, seed=1234):
    """Lehmer random number generator, following MINSTD.

    Reference: Commun. ACM. 12, 85.

    Args:
        size (tuple): Dimension of the array to create.

    Keyword Args:
        seed (int): Seed to initialize the random number generator.

    Returns:
        ndarray: Array with (pseudo) random numbers.
    """
    W = np.zeros(size, dtype=complex)
    mult = 48271
    mod = (2**31) - 1
    x = (seed * mult + 1) % mod
    for i in range(size[0]):
        for j in range(size[1]):
            for k in range(size[2]):
                x = (x * mult + 1) % mod
                W[i, j, k] = x / mod
    return W


def add_maybe_none(a, b):
    """Add a and b together, when one or both can potentially be None.

    Args:
        a (ndarray): Array.
        b (ndarray): Array.

    Returns:
        ndarray: Sum of a and b.
    """
    if a is b is None:
        return None
    if a is None:
        return b
    if b is None:
        return a
    return a + b


def molecule2list(molecule):
    """Expand a chemical formula to a list of chemical symbols.

    Args:
        molecule (str): Simplified chemical formula (case sensitive).

        No charges or parentheses are allowed, only chemical symbols followed by their amount.

    Returns:
        list: Atoms of the molecule expanded to a list.
    """
    # Insert a whitespace before every capital letter, these can appear once or none at all
    # Or insert before digits, these can appear at least once
    tmp_list = re.sub(r'([A-Z?]|\d+)', r' \1', molecule).split()
    atom_list = []
    for ia in tmp_list:
        if ia.isdigit():
            # If ia is an integer append the previous atom ia-1 times
            atom_list += [atom_list[-1]] * (int(ia) - 1)
        else:
            # If ia is a string add it to the results list
            atom_list += [ia]
    return atom_list


def atom2charge(atom, path=None):
    """Get the valence charges for a list of chemical symbols from GTH files.

    Args:
        atom (list): Atom symbols.
        path (str | None): Directory of GTH files.

    Returns:
        list: Valence charges per atom.
    """
    if path is not None:
        if path.lower() in ('pade', 'pbe'):
            psp_path = path.lower()
        else:
            psp_path = path
    else:
        psp_path = 'pbe'
    return [read_gth(ia, psp_path=psp_path)['Zion'] for ia in atom]



#!/usr/bin/env python3
"""Determine occupations for atomic systems from simple inputs."""
import dataclasses
import numbers

import numpy as np

from .logger import log


@dataclasses.dataclass
class Occupations:
    """Occupations class to save electronic state information in one place.

    The attribute Nelec has to be given first after instantiation.
    """
    # Set the private variable for the attributes that are properties.
    _Nelec: int = 0          #: Number of electrons.
    _Nspin: int = 0          #: Number of spin states.
    _spin: int = 0           #: Number of unpaired electrons.
    _charge: int = 0         #: System charge.
    _Nstate: int = 0         #: Number of states.
    is_filled: bool = False  #: Determines the Occupations object fill status.

    # ### Class properties ###

    @property
    def Nelec(self):
        """Number of electrons."""
        return self._Nelec

    @Nelec.setter
    def Nelec(self, value):
        # Only update Nelec if it actually gets updated
        if self._Nelec != int(value):
            self._Nelec = int(value)
            self.is_filled = False

    @property
    def Nspin(self):
        """Number of spin states."""
        return self._Nspin

    @Nspin.setter
    def Nspin(self, value):
        # Use a spin-paired calculation for an even number of electrons
        if value is None:
            if self.Nelec % 2 == 0 and self.spin == 0:
                value = 1
            else:
                value = 2
        # Only update if needed
        if self._Nspin != int(value):
            self._Nspin = int(value)
            self.spin = self.spin
            self.is_filled = False
            if hasattr(self, 'f'):
                log.warning('Reset previously set fillings.')

    @property
    def spin(self):
        """Number of unpaired electrons."""
        return self._spin

    @spin.setter
    def spin(self, value):
        # If no spin is given try to determine it from the number of electrons
        if value is None:
            if self.Nelec % 2 == 0:
                value = 0
            else:
                value = 1
        if self._spin != int(value):
            self.is_filled = False
        # We have no spin in the spin-paired case
        if self.Nspin == 1:
            self._spin = 0
        else:
            self._spin = int(value)

    @property
    def charge(self):
        """System charge."""
        return self._charge

    @charge.setter
    def charge(self, value):
        # If we set a charge via this setter update the number of electrons
        if hasattr(self, 'Nelec'):
            self.Nelec += self._charge - int(value)
        if self._charge != int(value):
            self._charge = int(value)
            self.is_filled = False

    @property
    def f(self):
        """Occupation numbers per state."""
        return self._f

    @f.setter
    def f(self, value):
        # Make sure the occupations are in a two-dimensional array
        if isinstance(value, (list, tuple, np.ndarray)):
            value = np.atleast_2d(value)
        self.is_filled = False
        # This setter will only be called when explicitly setting f
        # Call the fill function in that case
        self.fill(value)

    # ### Read-only properties ###

    @property
    def multiplicity(self):
        """Multiplicity, i.e., 2S+1."""
        return self.spin + 1

    @property
    def Nstate(self):
        """Number of states."""
        return self._Nstate

    @property
    def F(self):
        """Diagonal matrices of f per spin."""
        return [np.diag(f) for f in self.f]

    # ### Class methods ###

    def fill(self, f=None):
        """Fill the states of the object.

        Keyword Args:
            f: (float | ndarray | None): Fillings.
        """
        # Do nothing if the object is already filled
        if self.is_filled:
            return self
        # If no f is given just use the standard fillings: 2 for restricted and 1 for unrestricted
        if f is None:
            f = 2 / self.Nspin
        self._update_from_fillings(f)
        # Assure that no electrons have been lost
        if np.sum(self.f) != self.Nelec:
            ValueError(f'Sum of fillings ({np.sum(self.f)}) differs from Nelec ({self.Nelec}).')
        self.is_filled = True
        return self

    kernel = fill

    def _update_from_fillings(self, value):
        """Update fillings."""
        # Do not use the setter methods in this place to not trigger the setter effects
        # If f is a number use this occupation for all states
        if isinstance(value, numbers.Real):
            # Do not leave the states array empty when no electrons are present
            if self.Nelec <= 0:
                self._Nstate = 1
                self._f = np.zeros((self.Nspin, 1))
            elif self.Nspin == 1 or self.Nelec % 2 == self.spin % 2:
                self._integer_fillings(value)
            else:
                log.warning('Using fractional occupations.')
                self._fractional_fillings(value)
        # If f is an array redetermine all attributes from it
        else:
            self._f = value
            self._Nspin = len(value)
            self._Nstate = len(value[0])
            self._charge += self.Nelec - int(np.sum(value))
            self.Nelec = int(np.sum(value))
            if self.Nspin == 1:
                self.spin = 0
            else:
                self.spin = abs(np.sum(self.f[0] - self.f[1]))

    def _integer_fillings(self, f):
        """Update fillings while maintaining integer occupations numbers.

        Keyword Args:
            f: (int | ndarray): Fillings.
        """
        # Determine the electrons per spin channel
        Nup = self.Nelec // self.Nspin + self.spin // self.Nspin + self.Nelec % self.Nspin
        Ndw = self.Nelec // self.Nspin - self.spin // self.Nspin
        elecs = np.array([Nup, Ndw])
        # Get the number of states
        self._Nstate = int(np.ceil(max(elecs / f)))
        # Simply build the occupations array
        self._f = f * np.ones((self.Nspin, self._Nstate), dtype=int)
        # If we have filled too much correct it in the second spin channel
        # The following procedure ensures that we have no negative fillings
        rest = np.sum(self._f) - self.Nelec
        i = 1
        while rest > 0:
            # If the occupation is greater than the rest we are done with removing electrons
            if self._f[-1, -i] >= rest:
                self._f[-1, -i] -= rest
                break
            # Otherwise zero out the state, update the rest, and move to the next state
            rest -= self._f[-1, -i]
            self._f[-1, -i] = 0
            i += 1

    def _fractional_fillings(self, f):
        """Update fillings while allowing fractional occupation numbers.

        Keyword Args:
            f: (float | ndarray): Fillings.
        """
        # Determine the electrons per spin channel
        Nup = self.Nelec / self.Nspin + self.spin / self.Nspin
        Ndw = self.Nelec / self.Nspin - self.spin / self.Nspin
        elecs = np.array([Nup, Ndw])
        # Get the number of states
        self._Nstate = int(np.ceil(max(elecs / f)))
        # Simply build the occupations array
        self._f = f * np.ones((self.Nspin, self._Nstate))
        # If we have filled too much correct it in both spin channels
        # The following procedure ensures that we have no negative fillings
        rest = np.sum(self._f, axis=1) - elecs
        for s in range(self.Nspin):
            i = 1
            while rest[s] > 0:
                # If the occupation is greater than the rest we are done with removing electrons
                if self._f[s, -i] >= rest[s]:
                    self._f[s, -i] -= rest[s]
                    break
                # Otherwise zero out the state, update the rest, and move to the next state
                rest[s] -= self._f[s, -i]
                self._f[s, -i] = 0
                i += 1

    def __repr__(self):
        """Print the parameters stored in the Occupations object."""
        return f'Spin handling: {"un" if self.Nspin == 2 else ""}restricted\n' \
               f'Number of electrons: {self.Nelec}\n' \
               f'Spin: {self.spin}\n' \
               f'Charge: {self.charge}\n' \
               f'Number of states: {self._Nstate}\n' \
               f'Fillings: \n{self.f if self.is_filled else "Not filled"}'



#!/usr/bin/env python3
"""Setup file to make eminus installable.

For a full list of options see the documentation:
https://setuptools.pypa.io/en/latest/references/keywords.html
"""
import re
from typing import Dict, List

from setuptools import find_packages, setup

with open('eminus/version.py', 'r') as fh:
    version: str = \
        re.search(r"__version__ = '(.*?)'", fh.read()).group(1)  # type: ignore[union-attr]

with open('README.md', 'r') as readme, open('CHANGELOG.md', 'r') as changelog:
    long_description: str = readme.read() + '\n\n' + changelog.read().split('\n----\n')[0]

extras: Dict[str, List[str]] = {
    'dispersion': [
        'dftd3>=0.6.0'  # Interface for DFT-D3 dispersion corrections
    ],
    'libxc': [
        'pyscf>=1.7.3'  # Libxc interface via PySCF
    ],
    'torch': [
        'torch>=1.8'  # Faster FFT operators using Torch
    ],
    'viewer': [
        'matplotlib>=1.5.0',  # Plotting library
        'nglview>=2.6.5',     # Molecule and isosurface viewer
        'plotly>=4.5'         # Grid visualization
    ]
}
extras['fods'] = extras['libxc']  # PyCOM FOD guessing method uses PySCF
extras['all'] = [dep for values in extras.values() for dep in values]
extras['dev'] = [
    'coverage>=4.4',           # Generate coverage reports
    'furo>=2022.02.14.1',      # Documentation theme
    'mypy>=0.931',             # Static type checker
    'notebook',                # Run and convert notebooks to HTML
    'pytest>=5.4',             # Test utilities
    'pytest-cov>=2.6.1',       # Collect test coverage data
    'ruff>=0.0.276',           # Style guide checker
    'sphinx>=4',               # Documentation builder
    'sphinxcontrib-bibtex>=2'  # Use bib files for citations
]

setup(
    name='eminus',
    version=version,
    description='A plane wave density functional theory code.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Wanja Timm Schulze',
    author_email='wangenau@protonmail.com',
    url='https://github.com/wangenau/eminus',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development'
    ],
    license='APACHE2.0',
    keywords=['Python', 'DFT', 'DFT++'],
    include_package_data=True,
    install_requires=[
        'numpy>=1.17',
        'scipy>=1.6'
    ],
    extras_require=extras,
    python_requires='>=3.7',
    project_urls={
        'Bug Tracker': 'https://gitlab.com/wangenau/eminus/-/issues',
        'Changelog': 'https://wangenau.gitlab.io/eminus/changelog.html',
        'Documentation': 'https://wangenau.gitlab.io/eminus',
        'Source code': 'https://gitlab.com/wangenau/eminus'
    }
)



#!/usr/bin/env python3
"""Perdew-Burke-Ernzerhof GGA exchange for solids and surfaces.

Reference: Phys. Rev. Lett. 102, 039902.
"""
from .gga_x_pbe import gga_x_pbe, gga_x_pbe_spin


def gga_x_pbe_sol(n, **kwargs):
    """Perdew-Burke-Ernzerhof solid parametrization of the exchange functional (spin-paired).

    Corresponds to the functional with the label GGA_X_PBE_SOL and ID 116 in Libxc.

    Reference: Phys. Rev. Lett. 102, 039902.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: PBE exchange energy density, potential, and vsigma.
    """
    return gga_x_pbe(n, mu=10 / 81, **kwargs)


def gga_x_pbe_sol_spin(n, zeta, **kwargs):
    """Perdew-Burke-Ernzerhof solid parametrization of the exchange functional (spin-polarized).

    Corresponds to the functional with the label GGA_X_PBE_SOL and ID 116 in Libxc.

    Reference: Phys. Rev. Lett. 102, 039902.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray, ndarray]: PBE exchange energy density, potential, and vsigma.
    """
    return gga_x_pbe_spin(n, zeta, mu=10 / 81, **kwargs)



#!/usr/bin/env python3
"""Perdew-Wang LDA correlation.

Reference: Phys. Rev. B 45, 13244.
"""
import numpy as np


def lda_c_pw(n, A=0.031091, a1=0.2137, b1=7.5957, b2=3.5876, b3=1.6382, b4=0.49294, **kwargs):
    """Perdew-Wang parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_PW and ID 12 in Libxc.

    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        A (float): Functional parameter.
        a1 (float): Functional parameter.
        b1 (float): Functional parameter.
        b2 (float): Functional parameter.
        b3 (float): Functional parameter.
        b4 (float): Functional parameter.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    """
    rs = (3 / (4 * np.pi * n))**(1 / 3)
    rs12 = np.sqrt(rs)
    rs32 = rs * rs12
    rs2 = rs**2

    om = 2 * A * (b1 * rs12 + b2 * rs + b3 * rs32 + b4 * rs2)
    olog = np.log(1 + 1 / om)
    ec = -2 * A * (1 + a1 * rs) * olog

    dom = 2 * A * (0.5 * b1 * rs12 + b2 * rs + 1.5 * b3 * rs32 + 2 * b4 * rs2)
    vc = -2 * A * (1 + 2 / 3 * a1 * rs) * olog - 2 / 3 * A * (1 + a1 * rs) * dom / (om * (om + 1))
    return ec, np.array([vc]), None


def lda_c_pw_spin(n, zeta, A=(0.031091, 0.015545, 0.016887), fzeta0=1.709921, **kwargs):
    """Perdew-Wang parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_PW and ID 12 in Libxc.

    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        A (tuple): Functional parameters.
        fzeta0 (float): Functional parameter.
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    """
    a1 = (0.2137, 0.20548, 0.11125)
    b1 = (7.5957, 14.1189, 10.357)
    b2 = (3.5876, 6.1977, 3.6231)
    b3 = (1.6382, 3.3662, 0.88026)
    b4 = (0.49294, 0.62517, 0.49671)

    ec0, vc0, _ = lda_c_pw(n, A[0], a1[0], b1[0], b2[0], b3[0], b4[0])  # Unpolarized
    ec1, vc1, _ = lda_c_pw(n, A[1], a1[1], b1[1], b2[1], b3[1], b4[1])  # Polarized
    ac, dac, _ = lda_c_pw(n, A[2], a1[2], b1[2], b2[2], b3[2], b4[2])   # Spin stiffness
    ac = -ac  # The PW spin interpolation parametrizes -ac instead of ac

    fzeta = ((1 + zeta)**(4 / 3) + (1 - zeta)**(4 / 3) - 2) / (2**(4 / 3) - 2)
    zeta3 = zeta**3
    zeta4 = zeta3 * zeta

    ec = ec0 + ac * fzeta * (1 - zeta4) / fzeta0 + (ec1 - ec0) * fzeta * zeta4

    dac = -dac  # Also change the sign for the derivative
    dfzeta = ((1 + zeta)**(1 / 3) - (1 - zeta)**(1 / 3)) * 4 / (3 * (2**(4 / 3) - 2))
    factor1 = vc0 + dac * fzeta * (1 - zeta4) / fzeta0 + (vc1 - vc0) * fzeta * zeta4
    factor2 = (ac / fzeta0 * (dfzeta * (1 - zeta4) - 4 * fzeta * zeta3) +
               (ec1 - ec0) * (dfzeta * zeta4 + 4 * fzeta * zeta3))

    vc0p = factor1 + factor2 * (1 - zeta)
    vcdw = factor1 - factor2 * (1 + zeta)
    return ec, np.array([vc0p, vcdw]), None



#!/usr/bin/env python3
"""Fermi-orbital descriptor generation."""
import numpy as np
from scipy.linalg import norm

from ..data import SYMBOL2NUMBER
from ..logger import log
from ..units import bohr2ang


def get_localized_orbitals(mf, Nspin, loc, Nit=1000, seed=1234):
    """Generate localized orbitals with additional simple stability analysis.

    Same as implemented in PyFLOSIC2.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        mf: PySCF SCF object.
        Nspin (int): Number of spin states.
        loc (str): Localization method.

    Keyword Args:
        Nit (int): Number of tries to get a solution with positive eigenvalues.
        seed (int): Seed to initialize the random number generator.

    Returns:
        list: Localized occupied orbital coefficients per spin channel.
    """
    rng = np.random.default_rng(seed=seed)
    from pyscf.lo import boys, edmiston, pipek
    loc_dict = {'ER': edmiston.EdmistonRuedenberg,
                'FB': boys.Boys,
                'GPM': pipek.PipekMezey,
                'PM': pipek.PipekMezey}

    loc_orb = []
    # Localize each spin channel separately
    for s in range(Nspin):
        # Initialize the localizer object
        if Nspin == 2:
            localizer = loc_dict[loc](mf.mol, mf.mo_coeff[s][:, mf.mo_occ[s] > 0])
        else:
            localizer = loc_dict[loc](mf.mol, mf.mo_coeff[:, mf.mo_occ > 0])

        # Set the population method in generalized PM to Becke charges
        if loc == 'GPM':
            localizer.pop_method = 'becke'

        for _ in range(Nit):
            tmp_orb = localizer.kernel()
            # Calculate the eigenvalues of the Hessian
            _, _, h_diag = localizer.gen_g_hop(u=np.eye(len(tmp_orb[0])))
            # If there are no eigenvalues or all of them are positive break the loop
            if len(h_diag) == 0 or np.min(h_diag) > 0:
                break
            # If not continue with randomly 'displaced' orbitals
            noise = rng.normal(scale=5e-4, size=tmp_orb.shape)
            localizer.mo_coeff = tmp_orb + noise
        loc_orb.append(tmp_orb)
    return loc_orb


def get_fods(obj, basis='pc-1', loc='FB'):
    """Generate FOD positions using the PyCOM method.

    Reference: J. Comput. Chem. 40, 2843.

    Args:
        obj: Atoms or SCF object.

    Keyword Args:
        basis (str): Basis set for the DFT calculation.
        loc (str): Localization method.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        ndarray: FOD positions.
    """
    try:
        from pyscf.gto import Mole
        from pyscf.scf import RKS, UKS
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[fods]".\n\n')
        raise

    atoms = obj._atoms
    loc = loc.upper()

    # Convert to Angstrom for PySCF
    X = bohr2ang(atoms.X)
    # Build the PySCF input format
    atom_pyscf = list(zip(atoms.atom, X))

    # Do the PySCF DFT calculation
    # Use Mole.build() over M() since the parse_arg option breaks testing with pytest
    mol = Mole(atom=atom_pyscf, basis=basis, spin=atoms.spin).build(parse_arg=False)
    if atoms.unrestricted:
        mf = UKS(mol=mol)
    else:
        mf = RKS(mol=mol)
    mf.verbose = 0
    mf.kernel()

    # Get the localized orbital coefficients
    loc_orb = get_localized_orbitals(mf, atoms.occ.Nspin, loc)
    # Calculate the COMs
    loc_com = []
    ao = mf._numint.eval_ao(mf.mol, mf.grids.coords)
    for s in range(atoms.occ.Nspin):
        phi = ao @ loc_orb[s]
        dens = phi.conj() * phi * mf.grids.weights[:, None]
        loc_com.append(dens.T @ mf.grids.coords)
    return loc_com


def split_fods(atom, X, elec_symbols=('X', 'He')):
    """Split atom and FOD coordinates.

    Args:
        atom (list): Atom symbols.
        X (ndarray): Atom positions.

    Keyword Args:
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        tuple[list, ndarray, list]: Atom types, the respective coordinates, and FOD positions.
    """
    X_fod_up = []
    X_fod_dn = []
    # Iterate in reverse order because we may delete elements
    for ia in range(len(X) - 1, -1, -1):
        if atom[ia] in elec_symbols:
            if atom[ia] in elec_symbols[0]:
                X_fod_up.append(X[ia])
            if len(elec_symbols) > 1 and atom[ia] in elec_symbols[1]:
                X_fod_dn.append(X[ia])
            X = np.delete(X, ia, axis=0)
            del atom[ia]

    X_fod = [np.asarray(X_fod_up), np.asarray(X_fod_dn)]
    return atom, X, X_fod


def remove_core_fods(obj, fods):
    """Remove core FODs from a set of FOD coordinates.

    Args:
        obj: Atoms or SCF object.
        fods (list): FOD positions.

    Returns:
        ndarray: Valence FOD positions.
    """
    atoms = obj._atoms

    # If the number of valence electrons is the same as the number of FODs, do nothing
    if not atoms.unrestricted and len(fods[0]) * 2 == np.sum(atoms.occ.f[0]):
        return fods
    if atoms.unrestricted and len(fods[0]) == np.sum(atoms.occ.f[0]) \
            and len(fods[1]) == np.sum(atoms.occ.f[1]):
        return fods

    for s in range(atoms.occ.Nspin):
        for ia in range(atoms.Natoms):
            n_core = SYMBOL2NUMBER[atoms.atom[ia]] - atoms.Z[ia]
            # In the spin-paired case two electrons are one state
            # Since only core states are removed in pseudopotentials this value is divisible by 2
            # +1 to account for uneven amounts of core FODs (e.g., for hydrogen)
            n_core = (n_core + 1) // 2
            dist = norm(fods[s] - atoms.X[ia], axis=1)
            idx = np.argsort(dist)
            # Remove core FODs with the smallest distance to the core
            fods[s] = np.delete(fods[s], idx[:n_core], axis=0)
    return fods



#!/usr/bin/env python3
"""Utility functions for exchange-correlation functionals."""
import numpy as np

from .. import config
from ..logger import log
from ..utils import add_maybe_none
from .gga_c_chachiyo import gga_c_chachiyo, gga_c_chachiyo_spin
from .gga_c_pbe import gga_c_pbe, gga_c_pbe_spin
from .gga_c_pbe_sol import gga_c_pbe_sol, gga_c_pbe_sol_spin
from .gga_x_chachiyo import gga_x_chachiyo, gga_x_chachiyo_spin
from .gga_x_pbe import gga_x_pbe, gga_x_pbe_spin
from .gga_x_pbe_sol import gga_x_pbe_sol, gga_x_pbe_sol_spin
from .lda_c_chachiyo import lda_c_chachiyo, lda_c_chachiyo_spin
from .lda_c_chachiyo_mod import lda_c_chachiyo_mod, lda_c_chachiyo_mod_spin
from .lda_c_pw import lda_c_pw, lda_c_pw_spin
from .lda_c_pw_mod import lda_c_pw_mod, lda_c_pw_mod_spin
from .lda_c_vwn import lda_c_vwn, lda_c_vwn_spin
from .lda_x import lda_x, lda_x_spin


def get_xc(xc, n_spin, Nspin, dn_spin=None, tau=None, dens_threshold=0):
    """Handle and get exchange-correlation functionals.

    Args:
        xc (list | str): Exchange and correlation identifier.
        n_spin (ndarray): Real-space electronic densities per spin channel.
        Nspin (int): Number of spin states.

    Keyword Args:
        dn_spin (ndarray): Real-space gradient of densities per spin channel.
        tau (ndarray): Real-space kinetic energy densities per spin channel.
        dens_threshold (float): Do not treat densities smaller than the threshold.

    Returns:
        tuple[ndarray, ndarray]: Exchange-correlation energy density and potential.
    """
    if isinstance(xc, str):
        xc = parse_functionals(xc)
    f_exch, f_corr = xc

    # Only use non-zero values of the density
    n = np.sum(n_spin, axis=0)
    nz_mask = np.where(n > dens_threshold)
    n_nz = n[nz_mask]
    # Zeta is only needed for non-zero values of the density
    zeta_nz = get_zeta(n_spin[:, nz_mask])
    # dn_spin is only needed for non-zero values of the density
    if dn_spin is not None:
        dn_spin_nz = dn_spin[:, nz_mask[0], :]
    else:
        dn_spin_nz = None

    def handle_functional(fxc):
        """Calculate a given functional fxc, same for exchange and correlation."""
        # Calculate with the libxc extra...
        if ':' in fxc:
            from ..extras.libxc import libxc_functional
            fxc = fxc.split(':')[-1]
            exc, vxc, vsigma, vtau = libxc_functional(fxc, n_spin, Nspin, dn_spin, tau)
        # ...or use an internal functional
        else:
            if Nspin == 2 and fxc != 'mock_xc':
                fxc += '_spin'
            exc_nz, vxc_nz, vsigma_nz = IMPLEMENTED[fxc](n_nz, zeta=zeta_nz, dn_spin=dn_spin_nz,
                                                         Nspin=Nspin)
            # Map the non-zero values back to the right dimension
            exc = np.zeros_like(n)
            exc[nz_mask] = exc_nz
            vxc = np.zeros_like(n_spin)
            for s in range(Nspin):
                vxc[s, nz_mask] = vxc_nz[s]
            if vsigma_nz is not None:
                vsigma = np.zeros((len(vsigma_nz), len(exc)))
                for i in range(len(vsigma)):
                    vsigma[i, nz_mask] = vsigma_nz[i]
            else:
                vsigma = None
            # There are no internal meta-GGAs
            vtau = None
        return exc, vxc, vsigma, vtau

    ex, vx, vsigmax, vtaux = handle_functional(f_exch)  # Calculate the exchange part
    ec, vc, vsigmac, vtauc = handle_functional(f_corr)  # Calculate the correlation part
    return ex + ec, vx + vc, add_maybe_none(vsigmax, vsigmac), add_maybe_none(vtaux, vtauc)


def get_exc(*args, **kwargs):
    """Get the exchange-correlation energy density.

    This is a convenience function to interface :func:`~eminus.xc.utils.get_xc`.
    """
    exc, _, _, _ = get_xc(*args, **kwargs)
    return exc


def get_vxc(*args, **kwargs):
    """Get the exchange-correlation potential.

    This is a convenience function to interface :func:`~eminus.xc.utils.get_xc`.
    """
    _, vxc, vsigma, vtau = get_xc(*args, **kwargs)
    return vxc, vsigma, vtau


def parse_functionals(xc):
    """Parse exchange-correlation functional strings to the internal format.

    Args:
        xc (str): Exchange and correlation identifier, separated by a comma.

    Returns:
        list: Exchange and correlation string.
    """
    # Check for combined aliases
    try:
        # Remove underscores when looking up in the dictionary
        xc_ = xc.replace('_', '')
        xc = ALIAS[xc_]
    except KeyError:
        pass

    # Parse functionals
    functionals = []
    for f in xc.split(','):
        if ':' in f or f in IMPLEMENTED:
            f_xc = f
        elif not f:
            f_xc = 'mock_xc'
        else:
            try:
                # Remove underscores when looking up in the dictionary
                f_ = f.replace('_', '')
                f_xc = XC_MAP[f_]
            except KeyError:
                log.exception(f'No functional found for "{f}".')
                raise
        functionals.append(f_xc)

    # If only one or no functional has been parsed append with mock functionals
    functionals.extend('mock_xc' for _ in range(2 - len(functionals)))
    return functionals


def parse_xc_type(xc):
    """Parse functional strings to identify the corresponding functional type.

    Args:
        xc (list): Exchange and correlation identifier, separated by a comma.

    Returns:
        str: Functional type.
    """
    xc_type = []
    for func in xc:
        if ':' in func:
            xc_id = func.split(':')[-1]
            # Try to parse the functional using pylibxc at first
            try:
                family = parse_xc_libxc(xc_id)
            # Otherwise parse it with PySCF
            except (ImportError, AssertionError):
                family = parse_xc_pyscf(xc_id)

            if family == 1:
                xc_type.append('lda')
            elif family == 2:
                xc_type.append('gga')
            elif family == 4:
                xc_type.append('meta-gga')
            else:
                log.exception('Unsupported functional family.')
                raise
        # Fall back to internal xc functionals
        elif 'gga' in func:
            xc_type.append('gga')
        else:
            xc_type.append('lda')

    # When mixing functional types use the higher level of theory
    if xc_type[0] != xc_type[1]:
        log.warning('Detected mixing of different functional types.')
        if 'meta-gga' in xc_type:
            return 'meta-gga'
        return 'gga'
    return xc_type[0]


def parse_xc_libxc(xc_id):
    """Parse functional type by its ID using pylibxc.

    Args:
        xc_id (int | string): Functional ID or identifier.

    Returns:
        str: Functional type.
    """
    assert config.use_pylibxc
    import pylibxc
    if not xc_id.isdigit():
        xc_id = pylibxc.util.xc_functional_get_number(xc_id)

    func = pylibxc.LibXCFunctional(int(xc_id), 1)
    if func._needs_laplacian:
        log.exception('meta-GGAs that need a laplacian are not supported.')
        raise
    return func.get_family()


def parse_xc_pyscf(xc_id):
    """Parse functional type by its ID using PySCF.

    Args:
        xc_id (int | string): Functional ID or identifier.

    Returns:
        str: Functional type.
    """
    from pyscf.dft.libxc import is_gga, is_lda, is_meta_gga, needs_laplacian, XC_CODES
    if not xc_id.isdigit():
        xc_id = XC_CODES[xc_id.upper()]

    if needs_laplacian(int(xc_id)):
        log.exception('meta-GGAs that need a laplacian are not supported.')
        raise
    # Use the same values as in parse_xc_libxc
    if is_lda(xc_id):
        return 1
    if is_gga(xc_id):
        return 2
    if is_meta_gga(xc_id):
        return 4
    return -1


def get_zeta(n_spin):
    """Calculate the relative spin polarization.

    Args:
        n_spin (ndarray): Real-space electronic densities per spin channel.

    Returns:
        ndarray: Relative spin polarization.
    """
    # If only one spin is given return an array of ones as if the density only is in one channel
    if len(n_spin) == 1:
        return np.ones_like(n_spin[0])
    return (n_spin[0] - n_spin[1]) / (n_spin[0] + n_spin[1])


def mock_xc(n, Nspin=1, **kwargs):
    """Mock exchange-correlation functional with no effect (spin-paired).

    Args:
        n (ndarray): Real-space electronic density.
        Nspin (int): Number of spin states.

    Keyword Args:
        **kwargs: Throwaway arguments.

    Returns:
        tuple[ndarray, ndarray]: Mock exchange-correlation energy density and potential.
    """
    zeros = np.zeros_like(n)
    return zeros, np.array([zeros] * Nspin), None


#: Map functional names with their respective implementation.
IMPLEMENTED = {
    i.__name__: i for i in (
        mock_xc,
        gga_c_chachiyo,
        gga_c_chachiyo_spin,
        gga_c_pbe,
        gga_c_pbe_spin,
        gga_c_pbe_sol,
        gga_c_pbe_sol_spin,
        gga_x_chachiyo,
        gga_x_chachiyo_spin,
        gga_x_pbe,
        gga_x_pbe_spin,
        gga_x_pbe_sol,
        gga_x_pbe_sol_spin,
        lda_x,
        lda_x_spin,
        lda_c_pw,
        lda_c_pw_spin,
        lda_c_pw_mod,
        lda_c_pw_mod_spin,
        lda_c_vwn,
        lda_c_vwn_spin,
        lda_c_chachiyo,
        lda_c_chachiyo_spin,
        lda_c_chachiyo_mod,
        lda_c_chachiyo_mod_spin
    )
}

#: Map functional shorthands to the actual functional names.
XC_MAP = {
    # lda_x
    '1': 'lda_x',
    's': 'lda_x',
    'lda': 'lda_x',
    'slater': 'lda_x',
    # lda_c_vwn
    '7': 'lda_c_vwn',
    'vwn': 'lda_c_vwn',
    'vwn5': 'lda_c_vwn',
    # lda_c_pw
    '12': 'lda_c_pw',
    'pw': 'lda_c_pw',
    'pw92': 'lda_c_pw',
    # lda_c_pw_mod
    '13': 'lda_c_pw_mod',
    'pwmod': 'lda_c_pw_mod',
    'pw92mod': 'lda_c_pw_mod',
    # gga_x_pbe
    '101': 'gga_x_pbe',
    'pbex': 'gga_x_pbe',
    # gga_x_pbe_sol
    '116': 'gga_x_pbe_sol',
    'pbesolx': 'gga_x_pbe_sol',
    # gga_c_pbe
    '130': 'gga_c_pbe',
    'pbec': 'gga_c_pbe',
    # gga_c_pbe_sol
    '133': 'gga_c_pbe_sol',
    'pbesolc': 'gga_c_pbe_sol',
    # lda_c_chachiyo
    '287': 'lda_c_chachiyo',
    'chachiyo': 'lda_c_chachiyo',
    # gga_x_chachiyo
    '298': 'gga_x_chachiyo',
    'chachiyox': 'gga_x_chachiyo',
    # lda_c_chachiyo_mod
    '307': 'lda_c_chachiyo_mod',
    'chachiyomod': 'lda_c_chachiyo_mod',
    # gga_c_chachiyo
    '309': 'gga_c_chachiyo',
    'chachiyoc': 'gga_c_chachiyo'
}

#: Dictionary of common functional aliases.
ALIAS = {
    'svwn': 'slater,vwn5',
    'svwn5': 'slater,vwn5',
    'spw92': 'slater,pw92mod',
    'pbe': 'pbex,pbec',
    'pbesol': 'pbesolx,pbesolc',
    'chachiyo': 'chachiyox,chachiyoc'
}
